/* @(#)genxy1.c	14.1.1.1 (ESO-DMD) 09/16/99 09:28:18 */
/*===========================================================================
  Copyright (C) 1995 European Southern Observatory (ESO)
 
  This program is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public 
  License along with this program; if not, write to the Free 
  Software Foundation, Inc., 675 Massachusetts Ave, Cambridge, 
  MA 02139, USA.
 
  Correspondence concerning ESO-MIDAS should be addressed as follows:
	Internet e-mail: midas@eso.org
	Postal address: European Southern Observatory
			Data Management Division 
			Karl-Schwarzschild-Strasse 2
			D 85748 Garching bei Muenchen 
			GERMANY
===========================================================================*/

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.COPYRIGHT    (c)  1988   European Southern Observatory
.IDENT        genxy1.c
.LANGUAGE     C
.AUTHOR       K. Banse   ESO/IPG
.KEYWORDS     Bulk data frames, Images
.PURPOSE      rearrange planes of a cube
.VERSION      1.00  950921:   Creation 
990728
---------------------------------------------------------------------*/

#define NINT(a)      ((a) < 0 ? (int)((a) - 0.5 ) : (int)((a) + 0.5))

#define AsInt(ptr)   *(int *) ptr

#include   <midas_def.h>


static int OneFloat = 0x3F800000;

static float   factln, fact10, nullval, Scale, ScaleUp, ScaleDown;


/*

*/

int AsInteger(f)
float f;

{
return (*(int *) &f);
}




float AsFloat(i)
int  i;

{
return (*(float *) &i);
}

/*

*/

float Alog2(f)
float f;

{
float  mylog2;


mylog2 = (float) (AsInteger(f) - OneFloat);
mylog2 *= ScaleDown;
return (mylog2);
}



float Blog2(f)
float f;

{
register float  mylog2, nylog2;
float   Rlog2;


mylog2 = (float) (AsInteger(f) - OneFloat);
nylog2 = (float) (AsInteger(f*0.6666) - OneFloat);

Rlog2 = ScaleDown * 0.5 * (mylog2 + nylog2) + 0.344;
return (Rlog2);
}

/*

*/

float Asqrt(f)
float f;

{
int  myint;


myint = (AsInteger(f)>>1) + (OneFloat>>1);
return (AsFloat(myint));
}



float Bsqrt(f)
float f;

{
register float  mysqrt;
float   Rsqrt;


mysqrt = Asqrt(f);
Rsqrt = (mysqrt*mysqrt+f) / (2*mysqrt);

return (Rsqrt);
}

/*

*/

/*  this routine implements the ideas of Jim Blinn (now at Microsoft)
    as published in the magazine:
    IEEE Computer Graphics and Applications, July/August 1997, p. 80-84  */


void comp_as_jblinn(optio,ptra,ptrb,count)
int  optio;		/* 1 = log10, 2 = ln, 3 = sqrt */
char *ptra;		/* pointer to in/out array */
char *ptrb;		/* pointer to optional 2. in array */
int  count;		/* no. of data */

{
register int nr, iaux;
int  myint;

float  f;
register float  *fptr, *gptr;
register float  mysqrt, mylog2, ffa, ffb;


fptr = (float *) ptra;

if (optio == 1)				/* compute result = log10(f) */
   {
   ffa = Scale * fact10;
   ffb = 0.344 * fact10;
   iaux = OneFloat;

   for (nr=0; nr<count; nr++)
      {
      if (*fptr <= 0.0) 
         *fptr++ = nullval;
      else
         {
         mylog2 = (float) (AsInt(fptr) - iaux);
         *fptr *= 0.6666;
         mylog2 += (float) (AsInt(fptr) - iaux);
         *fptr++ = ffa*mylog2 + ffb;
         }
      }
   }

else if (optio == 2)			/* compute result = ln(f) */
   {
   ffa = Scale * factln;
   ffb = 0.344 * factln;
   iaux = OneFloat;

   for (nr=0; nr<count; nr++)
      {
      if (*fptr <= 0.0)
         *fptr++ = nullval;
      else
         {
         mylog2 = (float) (AsInt(fptr) - iaux);
         *fptr *= 0.6666;
         mylog2 += (float) (AsInt(fptr) - iaux);
         *fptr++ = ffa*mylog2 + ffb;
         }
      }
   }

else if (optio == 3)			/* compute result = sqrt(f) */
   {
   iaux = OneFloat>>1;
   for (nr=0; nr<count; nr++)
      {
      f = *fptr;
      if (f < 0.0)
         *fptr++ = nullval;
      else
         {
         myint = (AsInt(fptr)>>1) + iaux;
         mysqrt = *((float *) &myint); 			/* mysqrt = Asqrt(f) */
         *fptr++ = (mysqrt+(f/mysqrt)) * 0.5;
         }
      }
   }

else if (optio == 4)                    /* compute result = 1.0/f */
   {
   iaux = 2 * OneFloat;
   for (nr=0; nr<count; nr++)
      {
      f = *fptr;
      if ((f < -1.0e-8) || (f > 1.0e-8))
         {
         myint = iaux - AsInt(fptr);
         mysqrt = *((float *) &myint); 
         *fptr++ = mysqrt*(2-f*mysqrt);
         }
      else
         *fptr++ = nullval;
      }
   }

else if (optio == 5)                    /* compute result = f/g */
   {
   gptr = (float *) ptrb;
   iaux = 2 * OneFloat;
   for (nr=0; nr<count; nr++)
      {
      f = *gptr;
      if ((f < -0.1e-8) || (f > 0.1e-8))
         {
         myint = iaux - AsInt(gptr);
         mysqrt = *((float *) &myint);
         *fptr++ *= mysqrt*(2-f*mysqrt);
         }
      else
         *fptr++ = nullval;
      }
   }
}

/*

*/

void fast_comp(infile,outfile)
char    *infile, *outfile;

{
char    cbuf[12], in2file[64], *pntra, *pntrb;

int  naxis, npix[6], imnoa, imnob, imnoc, imnow1, imnow2;
int  felm, mapsize, size;
int  merk, null, nval, unit;
int  n, direc;

float   e_as_const = 2.71828182839999988;


ScaleUp = (float) (0x00800000);
ScaleDown = 1.0/ScaleUp;
Scale = 0.5 / ScaleUp;

fact10 = 1.0/Blog2(10.0);
factln = 1.0/Blog2(e_as_const);

(void) SCKRDR("NULL",2,1,&nval,&nullval,&unit,&null);  /* get user Null */

merk = CGN_INDEXC(infile,',');
if (merk > 0)
   {
   infile[merk++] = '\0';
   (void) strcpy(in2file,&infile[merk]);
   }

(void) SCKGETC("P3",1,10,&nval,cbuf);		/* get conversion option */
CGN_UPSTR(cbuf);

(void) SCKRDI("MONITPAR",20,1,&nval,&n,&unit,&null);  /* get mapping size */
mapsize = n * n;

if (cbuf[1] == 'O')				/* LOG10 */
   direc = 1;
else if (cbuf[0] == 'I')			/* INVERSE */
   direc = 4;
else if (cbuf[1] == 'N')			/* LN */
   direc = 2;
else if (cbuf[1] == 'Q')			/* SQRT */
   direc = 3;
else if (cbuf[0] == 'D')			/* DIV */
   {
   if (merk < 2) SCETER(3,"we need 2 inframes for DIV...");

   direc = 5;
   }


(void) SCFOPN(infile,D_R4_FORMAT,0,F_IMA_TYPE,&imnoa);
(void) SCDRDI(imnoa,"NAXIS",1,1,&nval,&naxis,&unit,&null);
if (naxis > 6) SCETER(3,"More than 6 dimensions...");

(void) SCDRDI(imnoa,"NPIX",1,naxis,&nval,npix,&unit,&null);

size = 1;				/*  get total size of frame  */
for (n=0; n< naxis; n++) size *= npix[n];
(void) SCFCRE(outfile,D_R4_FORMAT,F_O_MODE,F_IMA_TYPE,size,&imnoc);


/* allocate virtual memory for in/out frame buffers with right data types */

if (size < mapsize) mapsize = size;
(void) SCFCRE("work1",D_R4_FORMAT,F_X_MODE,F_IMA_TYPE,mapsize,&imnow1);
(void) SCFMAP(imnow1,F_X_MODE,1,mapsize,&nval,&pntra);

if (direc == 5)
   {
   int  mpix[6];

   (void) SCFOPN(in2file,D_R4_FORMAT,0,F_IMA_TYPE,&imnob);
   (void) SCDRDI(imnob,"NAXIS",1,1,&nval,&n,&unit,&null);
   if (naxis != n) SCETER(4,"Both inframes need same dimensions...");

   (void) SCDRDI(imnob,"NPIX",1,naxis,&nval,mpix,&unit,&null);
   for (n=0; n<naxis; n++)
      {
      if (npix[n] != mpix[n]) SCETER(4,"Both inframes need same NOPIX...");
      }
   (void) SCFCRE("work2",D_R4_FORMAT,F_X_MODE,F_IMA_TYPE,mapsize,&imnow2);
   (void) SCFMAP(imnow2,F_X_MODE,1,mapsize,&nval,&pntrb);
   }
else
   pntrb = pntra;


/*  now we do the calculations  */

felm = 0;
while (felm < size)
   {
   n = felm + 1;
   (void) SCFGET(imnoa,n,mapsize,&nval,pntra);
   if (direc == 5) 
      (void) SCFGET(imnob,n,mapsize,&nval,pntrb);
   comp_as_jblinn(direc,pntra,pntrb,nval);
   (void) SCFPUT(imnoc,n,nval,pntra);
   felm += nval;
   }

(void) SCDCOP(imnoa,imnoc,1,cbuf);      /* copy all descriptors */
CGN_DSCUPD(imnoc,imnoc," ");            /* update history */
}

/*

*/

void copyii(infile,outfile)
char    *infile, *outfile;

{
char    cbuf[12], dely[8];
char    *pntra, *pntrb;

int  naxis, npix[6], imnoa, imnob, imnow1, imnow2;
int  felm, mapsize, size;
int  info[5], null, nval, unit;
int  n, fmtin, fmtout, uflow, oflow;

float   cuts[4];


(void) SCKGETC("P3",1,10,&nval,cbuf); 			/* get output format */
(void) SCKGETC("P4",1,7,&nval,dely);  			/* get delete flag   */

(void) SCKRDI("MONITPAR",20,1,&nval,&n,&unit,&null);  /* get mapping size */
mapsize = n * n;

(void) SCFINF(infile,2,info);			/* get source format  */
fmtin =  info[1];
fmtout = check_fmt(cbuf);
if (fmtout < 0) SCETER(2,"invalid data format...");

(void) SCFOPN(infile,info[1],0,F_IMA_TYPE,&imnoa);
(void) SCDRDI(imnoa,"NAXIS",1,1,&nval,&naxis,&unit,&null);
if (naxis > 6) SCETER(3,"More than 6 dimensions...");

(void) SCDRDI(imnoa,"NPIX",1,naxis,&nval,npix,&unit,&null);
(void) SCDRDR(imnoa,"LHCUTS",1,4,&nval,cuts,&unit,&null);

uflow = 0; oflow = 0;
size = 1;					/*  get total size of frame  */
for (n=0; n< naxis; n++)
   size *= npix[n];

(void) SCFCRE(outfile,fmtout,F_O_MODE,F_IMA_TYPE,size,&imnob);


/* allocate virtual memory for in/out frame buffers with right data types */

if (size <= mapsize) mapsize = size;
(void) SCFCRE("work1",fmtin,F_X_MODE,F_IMA_TYPE,mapsize,&imnow1);
(void) SCFMAP(imnow1,F_X_MODE,1,mapsize,&nval,&pntra);

if (fmtin != fmtout)
   {
   (void) SCFCRE("work2",fmtout,F_X_MODE,F_IMA_TYPE,mapsize,&imnow2);
   (void) SCFMAP(imnow2,F_X_MODE,1,mapsize,&nval,&pntrb);
   }
else
   pntrb = pntra;


/*  now we do the conversion   */

felm = 0;
while (felm < size)
   {
   n = felm + 1;
   (void) SCFGET(imnoa,n,mapsize,&nval,pntra);
   if (fmtin != fmtout) 
      conv_pix(pntrb,pntra,fmtout,fmtin,nval);
   (void) SCFPUT(imnob,n,nval,pntrb);
   felm += nval;
   }


/*  handle over, underflow  */

if (fmtout == D_I1_FORMAT)
   {
   if (cuts[2] < 0) uflow = 1;
   if (cuts[3] > 255) oflow = 1;
   }
else if (fmtout == D_I2_FORMAT)
   {
   if (cuts[2] < -32768) uflow = 1;
   if (cuts[3] > 32767) oflow = 1;
   }
else if (fmtout == D_UI2_FORMAT)
   {
   if (cuts[2] < 0) uflow = 1;
   if (cuts[3] > 65534) oflow = 1;
   }
if (uflow == 1)
   SCTPUT("Warning: Format conversion results in underflow...");
if (oflow == 1)
   SCTPUT("Warning: Format conversion results in overflow...");


(void) SCDCOP(imnoa,imnob,1,cbuf);	/* copy all descriptors */

n = CGN_INDEXC(dely,',');		/* test history_update_flag */
if (n > 0)
   dely[1] = dely[n+1];
else
   dely[1] = 'u';

if ((dely[1] == 'U') || (dely[1] == 'u')) 
   (void) CGN_DSCUPD(imnob,imnob," ");		/* update history */



/* if delete flag set, delete the input frame  */

if ((dely[0] == 'D') || (dely[0] == 'd')) (void) SCFDEL(infile);
}

/*

*/

check_fmt(cbuf)
char  *cbuf;

{
int  ibra;

ibra = *cbuf;
cbuf ++;

switch (ibra)
   {
   case 'I':
   case 'i':

   if (*cbuf == '1')
      return (D_I1_FORMAT);
   else if (*cbuf == '2')
      return (D_I2_FORMAT);
   else if (*cbuf == '4')
      return (D_I4_FORMAT);
   else
      break;


   case 'U':
   case 'u':

   if ((*cbuf == 'I') || (*cbuf == 'i'))
      {
      if (*(++cbuf) == '2')
         return (D_UI2_FORMAT);
      }


   case 'R':
   case 'r':

   if (*cbuf == '8')
      return (D_R8_FORMAT);
   else
      return (D_R4_FORMAT);
   

   case 'D':
   case 'd':

   return (D_R8_FORMAT);


   default:
      break;
   }

return (-1);			/* wrong format  */
}

/*

*/

void transcube(infile,outfile)
char  *infile, *outfile;

{
char   cbuf[12], temp[128];
char   *pntra;

static char workfile[] = "middummxyz.bdf";

int  nobytes, mapsize, unit, null, fmtin;
int  naxis, npix[3], newpix[3], imnoa, imnob, imnow, size;
int  info[5], planedir, nval, mm[3];
register int  nr, mr;

double  start[3], step[3], nstart[3], nstep[3];


(void) SCFINF(infile,2,info);			/* get source format  */
nobytes = info[0];				/* bytes per pixel */
fmtin =  info[1];

(void) SCFOPN(infile,fmtin,0,F_IMA_TYPE,&imnoa);
(void) SCDRDI(imnoa,"NAXIS",1,1,&nval,&naxis,&unit,&null);
if (naxis != 3)
   SCETER(3,"Input image has not 3 dimensions...");
(void) SCDRDI(imnoa,"NPIX",1,naxis,&nval,npix,&unit,&null);
(void) SCDRDD(imnoa,"START",1,naxis,&nval,start,&unit,&null);
(void) SCDRDD(imnoa,"STEP",1,naxis,&nval,step,&unit,&null);


/* create working frame (may be renamed to output frame) */

size = npix[0] * npix[1] * npix[2];		/*  get total size of frame */
(void) SCFCRE(workfile,fmtin,F_O_MODE,F_IMA_TYPE,size,&imnob);


(void) SCKGETC("P3",1,4,&nval,cbuf); 			/* get plane specs */
CGN_UPSTR(cbuf);

if ( (strncmp(cbuf,"ZY",2) == 0) || (strncmp(cbuf,"YZ",2) == 0) )
   {
   mm[0] = 2;
   mm[1] = 1;
   mm[2] = 0;
   planedir = Z_Y_PLANE;
   if (cbuf[0] == 'Y')
      {
      SCTPUT(
"Warning: YZ-plane is handled as ZY-plane, i.e. rows are in z-direction...");
      }
   }
else if ( (strncmp(cbuf,"ZX",2) == 0) || (strncmp(cbuf,"XZ",2) == 0) )
   {
   mm[0] = 0;
   mm[1] = 2;
   mm[2] = 1;
   planedir = X_Z_PLANE;
   if (cbuf[0] == 'Z')
      {
      SCTPUT(
"Warning: ZX-plane is handled as XZ-plane, i.e. rows are in x-direction...");
      }
   }
else
   SCETER(3,"Invalid plane specification...");

for (nr=0; nr<3; nr++)
   {
   newpix[nr] = npix[mm[nr]];
   nstart[nr] = start[mm[nr]];
   nstep[nr] = step[mm[nr]];
   }

(void) SCDCOP(imnoa,imnob,1,cbuf); 		/* copy all descriptors */
(void) SCDWRI(imnob,"NPIX",newpix,1,naxis,&unit);	/* update NPIX */
(void) SCDWRD(imnob,"START",nstart,1,naxis,&unit);	/* update START */
(void) SCDWRD(imnob,"STEP",nstep,1,naxis,&unit);	/* update STEP */


/* allocate virtual memory for buffer with right data types */

mapsize = newpix[0] * newpix[1];
(void) SCFCRE("work1",fmtin,F_X_MODE,F_IMA_TYPE,mapsize,&imnow);
(void) SCFMAP(imnow,F_X_MODE,1,mapsize,&nval,&pntra);


/* loop through YZ- or XZ_planes and store them as XY_plane into
   workframe                                                     */

for (nr=0; nr<newpix[2]; nr++)
   {
   mr = nr + 1;
   (void) SCPGET(imnoa,planedir,mr,pntra);
   (void) SCPPUT(imnob,X_Y_PLANE,mr,pntra);
   }

CGN_DSCUPD(imnoa,imnob," ");		/* update history */


/* now test if result should replace input frame or go to a new frame */

if (outfile[0] == '+')
   CGN_CLEANF(infile,0,temp,120,&nval,&null);
else
   CGN_CLEANF(outfile,0,temp,120,&nval,&null);
(void) SCFRNM(workfile,temp);

}

/*

*/

main()

{
char   infile[82], outfile[64], cbuf[12];

int    nval;


(void) SCSPRO("genxy1");            

(void) SCKGETC("IN_A",1,80,&nval,infile);  		/* get input frame   */
(void) SCKGETC("OUT_A",1,60,&nval,outfile);  		/* get output frame  */

(void) SCKGETC("ACTION",1,3,&nval,cbuf);		/* get action flag */

if (strcmp(cbuf,"TRA") == 0) 
   transcube(infile,outfile);

else if (strcmp(cbuf,"COP") == 0) 
   copyii(infile,outfile);

else if (strcmp(cbuf,"FAS") == 0) 
   fast_comp(infile,outfile);

(void) SCSEPI();
}
