/* @(#)tbcopyref.c	14.1.1.1 (ES0-DMD) 09/16/99 09:29:43 */
/*===========================================================================
  Copyright (C) 1995 European Southern Observatory (ESO)
 
  This program is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public 
  License along with this program; if not, write to the Free 
  Software Foundation, Inc., 675 Massachusetss Ave, Cambridge, 
  MA 02139, USA.
 
  Corresponding concerning ESO-MIDAS should be addressed as follows:
	Internet e-mail: midas@eso.org
	Postal address: European Southern Observatory
			Data Management Division 
			Karl-Schwarzschild-Strasse 2
			D 85748 Garching bei Muenchen 
			GERMANY
===========================================================================*/

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.TYPE        Module
.NAME        tbcopyref.c
.LANGUAGE    C
.AUTHOR      IPG-ESO Garching
.CATEGORY    table utilities
.RETURNS
	status
.PURPOSE    

Copy Column from one table to another by reference.
The current version is coded using the standard table routines.
The code has to be optimized to a turbo version avoiding some of the checking.

Check for selected values.

.VERSION     1.0 25-Mar-1989   Definition     J.D. Ponz
------------------------------------------------------------*/
#include <midas_def.h>
#include <tbldef.h>

#define PARLEN 80

tbl_copyref(in,icol,iref,out,ocol,oref,type, nrow, nout)

int in		/* IN : input table id */;
int icol	/* IN : input column */;
int iref	/* IN : input reference column */;
int out		/* IN : output table id */;
int ocol	/* IN : output column */;
int oref	/* IN : output reference column */;
int type	/* IN : column type */;
int nrow	/* IN : total number of rows in input table */;
int *nout	/* OUT: actual numberof rows copied */;

{ int i, j, null, ival, jref, jerr, status, rtype, len, next, sel;
  float rval, rref, rerr;
  double dval, dref, derr;
  char   form[10], cval[256], cref[256];

  j = 0; status = 0; jerr = 0; rerr = 0; derr = 0; *nout = 0;

  				/* search reference data type */
  TCFGET(in,iref,form,&len,&rtype);

  switch (rtype) {
  case D_I1_FORMAT:
  case D_I2_FORMAT:
  case D_I4_FORMAT: 
       for (i = 1; i<=nrow; i++)
           {TCERDI(in,i,iref,&jref,&null);
            TCSGET(in,i,&sel);
            if ((!null) && sel) 
               {TCESRI(out,oref,jref,jerr,1,&next);
                if (next>0)
                   {*nout = *nout + 1;
                    switch (type) {
                    case D_I1_FORMAT:
                    case D_I2_FORMAT:
                    case D_I4_FORMAT:TCERDI(in,i,icol,&ival,&null);
                                     if (null)  TCEDEL(out,next,ocol);
                                     else TCEWRI(out,next,ocol,&ival);
                                     break;
                    case D_R4_FORMAT:TCERDR(in,i,icol,&rval,&null);
                                     if (null)  TCEDEL(out,next,ocol);
                                     else TCEWRR(out,next,ocol,&rval);
                                     break;
                    case D_R8_FORMAT:TCERDD(in,i,icol,&dval,&null);
                                     if (null)  TCEDEL(out,next,ocol);
                                     else TCEWRD(out,next,ocol,&dval);
                                     break;
                    case D_C_FORMAT: TCERDC(in,i,icol,cval,&null);
                                     if (null) TCEDEL(out,next,ocol);
                                     else TCEWRC(out,next,ocol,cval);
                                     break;
                    }                                    
                 }
             }
           }
       break;                              
  case D_R4_FORMAT: 
       for (i = 1; i<=nrow; i++)
           {TCERDR(in,i,iref,&rref,&null);
            TCSGET(in,i,&sel);
            if ((!null) && sel) 
               {TCESRR(out,oref,rref,rerr,1,&next);
                if (next>0)
                   {*nout = *nout + 1;
                    switch (type) {
                    case D_I1_FORMAT:
                    case D_I2_FORMAT:
                    case D_I4_FORMAT:TCERDI(in,i,icol,&ival,&null);
                                     if (null)  TCEDEL(out,next,ocol);
                                     else TCEWRI(out,next,ocol,&ival);
                                     break;
                    case D_R4_FORMAT:TCERDR(in,i,icol,&rval,&null);
                                     if (null)  TCEDEL(out,next,ocol);
                                     else TCEWRR(out,next,ocol,&rval);
                                     break;
                    case D_R8_FORMAT:TCERDD(in,i,icol,&dval,&null);
                                     if (null)  TCEDEL(out,next,ocol);
                                     else TCEWRD(out,next,ocol,&dval);
                                     break;
                    case D_C_FORMAT: TCERDC(in,i,icol,cval,&null);
                                     if (null) TCEDEL(out,next,ocol);
                                     else TCEWRC(out,next,ocol,cval);
                                     break;
                    }                                    
                }
             }
          }
       break;                              
  case D_R8_FORMAT: 
       for (i = 1; i<=nrow; i++)
           {TCERDR(in,i,iref,&dref,&null);
            TCSGET(in,i,&sel);
            if ((!null) && sel) 
               {TCESRD(out,oref,dref,derr,1,&next);
                if (next>0)
                   {*nout = *nout + 1;
                    switch (type) {
                    case D_I1_FORMAT:
                    case D_I2_FORMAT:
                    case D_I4_FORMAT:TCERDI(in,i,icol,&ival,&null);
                                     if (null) TCEDEL(out,next,ocol);
                                     else TCEWRI(out,next,ocol,&ival);
                                     break;
                    case D_R4_FORMAT:TCERDR(in,i,icol,&rval,&null);
                                     if (null) TCEDEL(out,next,ocol);
                                     else TCEWRR(out,next,ocol,&rval);
                                     break;
                    case D_R8_FORMAT:TCERDD(in,i,icol,&dval,&null);
                                     if (null) TCEDEL(out,next,ocol);
                                     else TCEWRD(out,next,ocol,&dval);
                                     break;
                    case D_C_FORMAT: TCERDC(in,i,icol,cval,&null);
                                     if (null) TCEDEL(out,next,ocol);
                                     else TCEWRC(out,next,ocol,cval);
                                     break;
                    }                                    
                  }
              }
         }
       break;                              
  case D_C_FORMAT: 
       for (i = 1; i<=nrow; i++)
           {TCERDC(in,i,iref,cref,&null);
            TCSGET(in,i,&sel);
            if ((!null) && sel) 
               {TCESRC(out,oref,cref,1,len,1,&next);
                if (next>0)
                   {*nout = *nout + 1;
                    switch (type) {
                    case D_I1_FORMAT:
                    case D_I2_FORMAT:
                    case D_I4_FORMAT:TCERDI(in,i,icol,&ival,&null);
                                               if (null) TCEDEL(out,next,ocol);
                                               else TCEWRI(out,next,ocol,&ival);
                                               break;
                    case D_R4_FORMAT:TCERDR(in,i,icol,&rval,&null);
                                               if (null) TCEDEL(out,next,ocol);
                                               else TCEWRR(out,next,ocol,&rval);
                                               break;
                    case D_R8_FORMAT:TCERDD(in,i,icol,&dval,&null);
                                               if (null) TCEDEL(out,next,ocol);
                                               else TCEWRD(out,next,ocol,&dval);
                                               break;
                    case D_C_FORMAT: TCERDC(in,i,icol,cval,&null);
                                               if (null) TCEDEL(out,next,ocol);
                                               else TCEWRC(out,next,ocol,cval);
                                               break;
                    }                                    
                 }
              }
           }
           break;                              
   }  


 return (status);
}       
